; /************************************************************************************
; |
; | 1on8-N5(2) - 15-Jan-2011
; | Node address 5 (0000 0101)
; |
; | This file is intended for 8 pin 12F675 PIC
; |
; |
; | 		+VDD     1+------+8  Gnd(Vss)
; | ErrLed	GP5(I/O) 2|      |7  GP0 (I/O) Tx
; | Output	GP4(I/O) 3|      |6  GP1 (I/O) DEn
; | Rx		GP3(I)   4+------+5  GP2 (I/O) InputSwitch
; |
; |	GP3 = Input ONLY
; |
; |------------------------------------------------------------------------------
; | GPIO3[4] ICSP - Vpp
; | GPIO1[6] ICSP - PGC (Clock)
; | GPIO0[7] ICSP - PGD (Data)
; | ------------------------------------------------------------------------------
; |
; |	Vss[8] - 0V(Gnd)
; |	GP0[7] - [O] - (ICSP/PGD) Tx
; |	GP1[6] - [O] - (ICSP/PGC) DEn
; |	GP2[5] - [I] - Switch Input
; |	GP3[4] - [I] - MCLR(ICSP/VPP) Rx 
; |	GP4[3] - [O] - Relay Output
; |	GP5[2] - [O] - RxErr
; |	Vdd[1] - +5V
; |
; \************************************************************************************

	LIST            P=12F675
	include "p12F675.inc"	; Refer to the *.inc file
	ERRORLEVEL      -302

		__CONFIG 0x0194
		; I want WatchDog timer OFF,
		; I want Internal RC no clock out
		; Brown out detect OFF
		; and I want Code Protect OFF

; - - - - - - - - - - - VARIABLE DECLARATION AREA - - - - - - - -
; GPIO,0 = ICSP Data
; GPIO,1 = ICSP Clock
; GPIO,3 = MCLR/Vpp

IOTris		equ		B'00001100'	;
#define 	Tx		GPIO,0		; <7>[0] Transmit <485 Tx>
#define 	DEn		GPIO,1		; <6>[0] DataEnable <485 DEn>
#define 	Inp		GPIO,2		; <5>[1] Input Switch 
#define 	Rx		GPIO,3		; <4>[1] GP3/MCLR <485 Rx>
#define		Outp		GPIO,4		; <3>[0] Output relay
#define		ErrLED		GPIO,5		; <2>[0] Rx Err LED
; ------------------------------------------------------------------------------

;-------------------------------------------------------------------
; Set Node Address. 
; Each Receiver unit should have a unique address set here
;-------------------------------------------------------------------
	CBLOCK  20H          	  		; define variables required
			count			; used in looping routines
			count1			; used in delay routine
			counta			; used in delay routine
			countb			; used in delay routine
			TX_Byte			; holds byte to xmit
			RX_Byte			; holds received byte 
			Bit_Cntr		; bit counter for RS232
			Temp			; Register for Temp use
			Delay_Count        	; delay loop counter
			SwReg			; Switch holding register
			NodeNo			; Node address
		endc

; --- End section variable declarations --------------------------------------- ;

; ===================================================== ;
							;
		org		0x00			;
		call		Start			;
		goto		Loop			;
							;
; ===================================================== ;
		org 		0x04 			; Jump location for interrupts
		retfie					;
; ===================================================== ;

; --- Loop -------------------------------------------- ;
Loop							;
		call		RX_232			;
		call		ChkSw			; Map the switch positions into switch register
		call		AddNode			; Put NodeNo and SwReg data into TX_Byte
		call		Decode			; Write Received data to LED's
		btfsc		ErrLED			; Test if the receive addr did match the node no?
		goto		Loop			; No, do not transmit, go back
;		bcf		NodeErr			;
		bcf		ErrLED			;		
		goto		Loop			;
; ----------------------------------------------------- ;


; --- Check Switches ---------------------------------------------------------- ;
; Weak Pull-up is enabled in the PIC. 
; If a switch is closed it will give logic 0. 
; If a Switch is Open it will give Logic 1.
; -----------------------------------------------------------------------------	;
ChkSw:							;

		btfsc		Inp			; Check Switch input
		bcf		SwReg,7			; Set On for Node1
		btfss		Inp			; Check Switch input
		bsf		SwReg,7			;
		return					;

; ===================================================== ;
AddNode:
; ----------------------------------------------------------------------------- ;
; The master talks to each node in turn. This code calculates
; the next node to be communicated with. Once Node 8 (0000111)
; is reached, the nodeno is reset to 0x00 (Node1)
; -----------------------------------------------------------------------------	;

; ----------------------------------------------------------------------------- ;
; We now combine the NodeNo with the Switch Register data into the TX_Byte
; ----------------------------------------------------------------------------- ;
		clrf		TX_Byte			; Clear TX_Byte
		movf		NodeNo,w		; put node no in w
		iorwf		TX_Byte,1		; put node no Tx_Byte
		movf		SwReg,w			; Move SwReg into w
		iorwf		TX_Byte,1		; to put it into TX_Byte
		return					;
; ----------------------------------------------------- ;		


; ===================================================== ;
;		Serial routines
; ----------------------------------------------------- ;
; --- Serial Transmit --------------------------------------------------------- ; 
;
; For RS232 we need to have a start bit at logic level 0		
; This means that the Tx line is always kept high during idle
; conditions. For RS485 half duplex communication this poses a problem.
; We cannot keep the Tx high and then drop because the RS485 lines do both
; Transmit and receive.
;
; (Transmit routine)	
;      St  0   1   2   3   4   5   6   7   Sp
;    +---+---+---+---+---+---+---+---+---+---+
;    | 1 |/X/|/X/|/X/|/X/|/X/|/X/|/X/|/X/| 1 |
;----+   +---+---+---+---+---+---+---+---+   +---------
;      |   |   |   |   |   |   |   |   |   |
;      |   |   |   |   |   |   |   |   |   +- Stop Bit Logic 1
;      |   |   |   |   |   |   |   |   +----- MSB Nibble <7>
;      |   |   |   |   |   |   |   +---------     Nibble <6>
;      |   |   |   |   |   |   +-------------     Nibble <5>
;      |   |   |   |   |   +-----------------     Nibble <4>
;      |   |   |   |   +---------------------     Nibble <3>
;      |   |   |   +-------------------------     Nibble <2>
;      |   |   +-----------------------------     Nibble <1>
;      |   +--------------------------------- LSB Nibble <0>
;      +------------------------------------- Start Bit Logic 1
;
;
; The TX cycle should take a total of 1ms to complete

; ----------------------------------------------------------------------------- ; 

XMIT_RS232	
		bsf		DEn			; Enable data transmit on RS485
		movlw		0x08			; set 8 bits out
		movwf		Bit_Cntr		; move to bit counter
		bsf		Tx			; Logic 1 for Start Bit - (Needed for RS485)
		call		Bit_Delay		; 100us delat
Ser_Loop	rrf		TX_Byte , f		; Rotate through C to test for next bit
		btfss		STATUS    , C		; If C is off
		bcf		Tx			; TX Off
		btfsc		STATUS    , C		; If C is ON
		bsf		Tx			; TX ON
		call		Bit_Delay		; 100us delay
		decfsz		Bit_Cntr  , f		; test if all 8 bits are done
		goto		Ser_Loop		; No, next bit
		bsf		Tx			; Yes, switch On TX (Stop Bit)
		call		Bit_Delay		; 100us
		bcf		Tx			; Switch OFF TX
		bcf		DEn			; Switch OFF DataEnable
		return

; --- Serial Receive ---------------------------------------------------------- ;
; Receive data routine
; Should no data be received within a period of 5ms (Timer1 timeout)
; the routine is returned
; ----------------------------------------------------------------------------- ;

; (Receive routine)
;      St  0   1   2   3   4   5   6   7   Sp
;    +---+---+---+---+---+---+---+---+---+---+
;    | 1 |/X/|/X/|/X/|/X/|/X/|/X/|/X/|/X/| 1 |
;----+   +---+---+---+---+---+---+---+---+   +---------
;      |   |   |   |   |   |   |   |   |   |
;      |   |   |   |   |   |   |   |   |   +- Stop Bit Logic 1
;      |   |   |   |   |   |   |   |   +----- MSB Nibble <7>
;      |   |   |   |   |   |   |   +---------     Nibble <6>
;      |   |   |   |   |   |   +-------------     Nibble <5>
;      |   |   |   |   |   +-----------------     Nibble <4>
;      |   |   |   |   +---------------------     Nibble <3>
;      |   |   |   +-------------------------     Nibble <2>
;      |   |   +-----------------------------     Nibble <1>
;      |   +--------------------------------- LSB Nibble <0>
;      +------------------------------------- Start Bit Logic 1
;
; ----------------------------------------------------------------------------- ;

RX_232:							;	
RxWait:		btfss		Rx			; Test if Rx is set <1>
		goto		RxWait			; NO
; ----------------------------------------------------- ;
		call		Start_Delay		; do half bit time delay (53us)
		btfss		Rx			; check still in start bit
		goto		RxWait			; NO
; ----------------------------------------------------- ;
; At this point we confirm that the start bit (0) has
; been received and we can disable Timer1 watchdog	
; ----------------------------------------------------- ;
		movlw		0x08			; set up to read 8 bits
		movwf		Bit_Cntr		;
		clrf		RX_Byte			;
Next_RcvBit	call		Bit_Delay		; 101us Delay
		btfss		Rx			;
		bcf		STATUS,C		;
		btfsc		Rx			;	
		bsf		STATUS,C		;
		rrf		RX_Byte,f		;
		decfsz		Bit_Cntr,f		; test if all done
		goto		Next_RcvBit		;
		call		Bit_Delay		; 101us Delay
		return					;

; --- Serial Receive ---------------------------------------------------------- ;
; Received data to be displayed on LED
; Nodes 1 to 4 each have one Input and one output signal
; The master transmit a request (On or Off) and then that drives the Slaves
; Output according the the requirements from the master. 
; The result of the aciton taken by the slave (feedback) is then sent back to 
; the master for confirmation.
; Received data is received in RX_Byte, where the first nibble (bits 0-3) is
; the node number of the slave node, and the last nibble (bits 4-7) the corresponding
; data. (i.e. Node 1's data is on bit<4>, Node 2's data is on bit<5> etc.
; ----------------------------------------------------------------------------- ;
Decode:


GetNode:	movf		RX_Byte,W		;
		andlw		0x0F			; filter out node address
		movwf		Temp			; Place into Temp reg
		subwf		NodeNo,w		;
		btfss		STATUS,Z		; test if the ZERO bit is set
		goto		DataErr			; Not the same
							;
N01:		btfss		RX_Byte,7		;
		bcf		Outp			;
		btfsc		RX_Byte,7		;
		bsf		Outp			;
		call		XMIT_RS232		; Transmite data in TX_Byte
		return					;
							;
DataErr							;
		bsf		ErrLED			; switch on Error LED
		return					;
							;
; ==============================================================================
; Miscellaneous routines
; ==============================================================================

; ===================================================== ;
; Initialise the PIC					;	
; --- Start ------------------------------------------- ;
Start
		bcf		STATUS,RP0		; select register bank 0
		bcf		INTCON,GIE		; Enable Global interrupts	
		clrf		GPIO			; Init GPIO
		movlw		07h			; set GP<2:0> to (Chapter 6.2 of datasheet)
		movwf		CMCON			; digital IO
		bsf     	STATUS,RP0		; BANK 1
		clrf		ANSEL			; Digital I/O
		movlw		IOTris			; Get ports settings
		movwf		TRISIO			; Set GP<0,2,4,5> as Output
		movlw		b'00000000'		; Weak pullup enabled
		movwf		OPTION_REG		;
		movlw		b'00000100'		; I Want GPIO<2>
		movwf		WPU			; Disable weak pull-up
;		call		OSCC			; Get OSCCAL value and permanent copy in code
		call		h'3FF'          	; You have to manually enter this value in the calibration input
        	movwf   	OSCCAL    	      	; h'3FF' is where the calibration value is stored. Usually 34xx
		bcf     	STATUS,RP0		; BANK 0
		call		ClrReg			;
; ####################################################	;
; --- Variable node address --------------------------	;
; ####################################################	;
							;
		movlw		0x05			; Set Node No
		movwf		NodeNo			; Fix Node No
		return					;
							;
; ===================================================== ;

; ----------------------------------------------------- ;
; Indirect address resetting
; ----------------------------------------------------- ;
ClrReg		movlw		0x20			; Put first address in for
		movwf		FSR			; FSR pointer to go to
Next:		clrf		INDF			; Clear the register at that address
		incf		FSR,f			; increment FSR with 1
		movf		FSR,w			; Put FSR in w
		sublw		0x30			; Clear all registers from 0x20-0x30
		btfss		STATUS,Z		; Test if all done
		goto		Next			; No
		return					;


; ==============================================================================
; NB!!! NB!!! NB!!! NB!!! NB!!! NB!!!
; When programming the PIC for the first time, read the PIC and store the 
; OSCCAL value stored in pos 0x3ff (1023) (Typical 34XX). Copy the XX value
; in here  ---> retlw	0xXX EG 3444 = retlw 0x44
; ##############################################################################
; When making use of PicKit2 programmer, first read the OSCC value of the PIC, or
; let the programmer re-calculate the timer value. Use the programmer to store
; the value in the PIC
; ------------------------------------------------------------------------------
OSCC	retlw		0x50		;
; ==============================================================================

; ----------------------------------------------------- ;
; Generates 50us Delay
; ----------------------------------------------------- ;
Start_Delay	movlw		d'010'			; Was 0x0C
		movwf		Delay_Count		;
Start_Wait	nop					;
		decfsz		Delay_Count , f		;
		goto		Start_Wait		;
		return					;

; ------------------------------------------------------
;
; ------------------------------------------------------
Bit_Delay	movlw		d'022'			; was 0x18
		movwf		Delay_Count		;
Bit_Wait	nop					;
		decfsz		Delay_Count , f		;
		goto		Bit_Wait		;
		return					;

; ------------------------------------------------------
;Delay routines
; ------------------------------------------------------
Long_Delay
		call		Delay255		;
		call		Delay255		;
		call		Delay255		;
		call		Delay255		;
		return					;

Delay255	movlw		0xff			;delay 255 mS
		goto		d0			;
Delay100	movlw		d'100'			;delay 100mS
		goto		d0			;
Delay50		movlw		d'50'			;delay 50mS
		goto		d0			;
Delay20		movlw		d'20'			;delay 20mS
		goto		d0			;
Delay10		movlw		d'10'			;delay 10mS
		goto		d0			;
Delay1		movlw		d'1'			;delay 1mS
		goto		d0			;
Delay5		movlw		0x05			;delay 5.000 ms (4 MHz clock)
d0		movwf		count1			;
d1		movlw		0xC7			;
		movwf		counta			;
		movlw		0x01			;
		movwf		countb			;
Delay_0		decfsz		counta, f		;
		goto		$+2			;
		decfsz		countb, f		;
		goto		Delay_0			;

		decfsz		count1	,f		;
		goto		d1			;
		retlw		0x00			;

; ===================================================== ;
; end of Miscellaneous routines
; ------------------------------------------------------

; ==============================================================================

Prop:		dt "1on8-N5(2).asm"
;/------------------------------;
;	END PROGRAM		;
;\------------------------------;
		end		;
; ------------------------------------------------------------------------------

CREDITS
I based the Rx/Tx routines on code from Nigel Goodwin's page
http://www.winpicprog.co.uk/pic_tutorial7.htm
http://www.winpicprog.co.uk/