; /**********************************************
; | 
; | RS485 Master. Controls up to 8 nodes
; | Master18(2) - Jan-2011
; |
; |
; |	PIC16F876A	
; |
; |		    +-----------+
; |	    Vpp/MCLR|01       28|RB7(PGD)
; |		 RA0|02       27|RB6(PGC)
; |		 RA1|03       26|RB5
; |		 RA2|04       25|RB4
; |		 RA3|05   P   24|RB3/PGM
; |		 RA4|06   I   23|RB2
; |		 RA5|07   C   22|RB1
; |	     (0V)Vss|08   F   21|RB0/INT
; |		OSC1|09   8   20|Vdd(+5V)   
; |		OSC2|10   7   19|(0V)Vss   
; |		 RC0|11   6   18|RC7   
; |		 RC1|12   A   17|RC6   
; |		 RC2|13       16|RC5   
; |		 RC3|14       15|RC4   
; |		    +-----------+
; |
; |	PORTA0[01] - [DI] - Rx
; |	PORTA1[03] - [DO] - DEn
; |	PORTA2[04] - [DO] - Tx
; |	PORTA3[05] - [DO] - Synch
; |	PORTA4[06] - [] - 
; |	PORTA5[07] - [] -
; |				
; |	PORTB0[21] - [DI] - Sw01
; |	PORTB1[22] - [DI] - Sw02
; |	PORTB2[23] - [DI] - Sw03
; |	PORTB3[24] - [DI] - Sw04
; |	PORTB4[25] - [DI] - Sw05
; |	PORTB5[26] - [DI] - Sw06
; |	PORTB6[27] - [DI] - Sw07 (ICSP/PGC)
; |	PORTB7[28] - [DI] - Sw08 (ICSP/PGD)
; |			
; |	PORTC0[11] - [DO] - LED01
; |	PORTC1[12] - [DO] - LED02
; |	PORTC2[13] - [DO] - LED03
; |	PORTC3[14] - [DO] - LED04
; |	PORTC4[15] - [DO] - LED05
; |	PORTC5[16] - [DO] - LED06
; |	PORTC6[17] - [DO] - LED07
; |	PORTC7[18] - [DO] - LED08
; |
; ============================================================================= ;
	LIST P=16F876A, R=DEC    ; Use the PIC16F876A and decimal system 
	ERRORLEVEL  -302, -305
	#include "P16F876A.INC"  ; Include header file 
    __CONFIG _CP_OFF & _WDT_OFF & _BODEN_ON & _PWRTE_ON & _XT_OSC & _WRT_OFF & _CPD_OFF & _LVP_OFF
     
; ============================================================================= ;

		cblock	0x20			; start of general purpose registers
			count			; used in looping routines
			count1			; used in delay routine
			counta			; used in delay routine
			countb			; used in delay routine
			TX_Byte			; holds byte to xmit
			RX_Byte			; holds received byte 
			Bit_Cntr		; bit counter for RS232
			Bit_Cntr2		;
			Temp_Byte		;
			Delay_Count        	; delay loop counter
			GenPurp			; General Purpose register
			SwReg			; Switch holding register
			NodeNo			; Node address
			RxNodeNo		; Receiving node number
		endc
; ============================================================================= ;

#define 	Sw01		PORTB,0			; [21] Switch 01
#define 	Sw02		PORTB,1			; [22] Switch 02
#define 	Sw03		PORTB,2			; [23] Switch 03
#define 	Sw04		PORTB,3			; [24] Switch 04
#define 	Sw05		PORTB,4			; [25] Switch 05
#define 	Sw06		PORTB,5			; [26] Switch 06
#define 	Sw07		PORTB,6			; [27] Switch 07
#define 	Sw08		PORTB,7			; [28] Switch 08

#define		Rx		PORTA,0			; [02] Rx - Serial Receive
#define		DEn		PORTA,1			; [03] DEn - Data Enable on RS485
#define		Tx		PORTA,2			; [04] Tx - Serial Transmit
#define		Synch		PORTA,3			; [05] Synch
; ----------------------------------------------------------------------------- ;
#define		ISRFlag		GenPurp,0		; ISR activated 100us done
#define		Tmr1Flag	GenPurp,1		; Timer1 ISR activated
; ----------------------------------------------------------------------------- ;
#define 	Val		.64537			; Timer1 - 1ms delay
; ----------------------------------------------------------------------------- ;

; --- End section variable declarations --------------------------------------- ;

; ===================================================== ;
							;
		org		0x00			;
		call		Start			;
		goto		Loop			;
							;
; ===================================================== ;
		org 		0x04 			; Jump location for interrupts
ISR:		bcf		PIR1,TMR1IF		; Reset Timer overflow flag (When interrupt occurred)
		bsf		Tmr1Flag		; Set timer 1 interrupt flag
		retfie					;
; ===================================================== ;

; --- Loop -------------------------------------------- ;
Loop							;
		call		ChkSw			; Map the switch positions into switch register
		call		AddNode			; Put NodeNo and SwReg data into TX_Byte
		call		XMIT_RS232		; Transmite data in TX_Byte
		call		RX_232			; Receive data in RX_Byte
		call		Decode			; Write Received data to LED's
		goto		Loop			; Repeat
; ----------------------------------------------------- ;


; ============================================================================= ;
;
; --- Check Switches ---------------------------------------------------------- ;
; Weak Pull-up is enabled in the PIC. 
; If a switch is closed it will give logic 0. 
; If a Switch is Open it will give Logic 1.
; -----------------------------------------------------------------------------	;

; ----------------------------------------------------------------------------- ;
; The master talks to each node in turn. This code calculates
; the next node to be communicated with. Once Node 8 (0000111)
; is reached, the NodeNo is reset to 0x00 (Node1)
; -----------------------------------------------------------------------------	;
ChkSw:
		incf		NodeNo,f		; NodeNo + 1 
		movf		NodeNo,w		; Get NodeNo address
		sublw		0x08			; Test if is the last node?
		btfsc		STATUS,Z		; test if the ZERO bit is set
		clrf		NodeNo			; Back to node0		
; ----------------------------------------------------- ;
		movf		NodeNo,w		;
		andlw		7			; w = index, 0..7
		addwf		PCL,f			; 
; ----------------------------------------------------- ;
		goto		T01			; Test Sw01
		goto		T02			; Test Sw02
		goto		T03			; Test Sw03
		goto		T04			; Test Sw04
		goto		T05			; Test Sw05
		goto		T06			; Test Sw06
		goto		T07			; Test Sw07
		goto		T08			; Test Sw08
		return					; No valid node
; ----------------------------------------------------- ;
T01:		btfss		Sw01			; Check Switch input
		bsf		SwReg,7			; Set On for Node1
		btfsc		Sw01			; Check Switch input
		bcf		SwReg,7			;
		return		
; ----------------------------------------------------- ;
T02:		btfss		Sw02			; Check Switch input
		bsf		SwReg,7			; Set On for Node2
		btfsc		Sw02			; Check Switch input
		bcf		SwReg,7			;
		return		
; ----------------------------------------------------- ;
T03:		btfss		Sw03			; Check Switch input
		bsf		SwReg,7			; Set On for Node3
		btfsc		Sw03			; Check Switch input
		bcf		SwReg,7			;
		return		
; ----------------------------------------------------- ;
T04:		btfss		Sw04			; Check Switch input
		bsf		SwReg,7			; Set On for Node4
		btfsc		Sw04			; Check Switch input
		bcf		SwReg,7			;
		return					;
; ----------------------------------------------------- ;
T05:		btfss		Sw05			; Check Switch input
		bsf		SwReg,7			; Set On for Node5
		btfsc		Sw05			; Check Switch input
		bcf		SwReg,7			;
		return					;
; ----------------------------------------------------- ;
T06:		btfss		Sw06			; Check Switch input
		bsf		SwReg,7			; Set On for Node6
		btfsc		Sw06			; Check Switch input
		bcf		SwReg,7			;
		return					;
; ----------------------------------------------------- ;
T07:		btfss		Sw07			; Check Switch input
		bsf		SwReg,7			; Set On for Node7
		btfsc		Sw07			; Check Switch input
		bcf		SwReg,7			;
		return					;
; ----------------------------------------------------- ;
T08:		btfss		Sw08			; Check Switch input
		bsf		SwReg,7			; Set On for Node8
		btfsc		Sw08			; Check Switch input
		bcf		SwReg,7			;
		return					;
; ----------------------------------------------------- ;

AddNode:
; ----------------------------------------------------------------------------- ;
; We now combine the NodeNo with the Switch Register data into the TX_Byte
; ----------------------------------------------------------------------------- ;
		clrf		TX_Byte			; Clear TX_Byte
		movf		NodeNo,w		; put node no in w
		iorwf		TX_Byte,1		; put node no Tx_Byte
		movf		SwReg,w			; Move SwReg into w
		iorwf		TX_Byte,1		; to put it into TX_Byte
		return					;
; ----------------------------------------------------- ;		

; ============================================================================= ;
; Serial routines - TX - TX - TX - TX - TX - TX - TX - TX - TX - TX - TX - TX
; ============================================================================= ;

; --- Serial Transmit --------------------------------------------------------- ; 
;
; For RS232 we need to have a start bit at logic level 0		
; This means that the Tx line is always kept high during idle
; conditions. For RS485 half duplex communication this poses a problem.
; We cannot keep the Tx high and then drop to zero, because the RS485 lines do both
; Transmit and receive. 
;
; (Transmit routine)	
;      St  0   1   2   3   4   5   6   7   Sp
;    +---+---+---+---+---+---+---+---+---+---+
;    | 1 |/X/|/X/|/X/|/X/|/X/|/X/|/X/|/X/| 1 |
;----+   +---+---+---+---+---+---+---+---+   +---------
;      |   |   |   |   |   |   |   |   |   |
;      |   |   |   |   |   |   |   |   |   +- Stop Bit Logic 1
;      |   |   |   |   |   |   |   |   +----- MSB Nibble <7> Node bit data
;      |   |   |   |   |   |   |   +---------     Nibble <6> Not used
;      |   |   |   |   |   |   +-------------     Nibble <5> Not used
;      |   |   |   |   |   +-----------------     Nibble <4> Not used
;      |   |   |   |   +---------------------     Nibble <3> Node 2^2
;      |   |   |   +-------------------------     Nibble <2> Node 2^1
;      |   |   +-----------------------------     Nibble <1> Node 2^0
;      |   +--------------------------------- LSB Nibble <0> 
;      +------------------------------------- Start Bit Logic 1
;
;
; The TX cycle should take a total of 1ms to complete
;
; ----------------------------------------------------------------------------- ; 

XMIT_RS232
		bsf		Synch			; Start Sycnh for Oscil
		bsf		DEn			; Enable data transmit on RS485
		movlw		0x08			; set 8 bits out
		movwf		Bit_Cntr		; move to bit counter
		bsf		Tx			; Logic 1 for Start Bit - (Needed for RS485)
		bcf		Synch			; Reset Oscil Synch
		call		Bit_Delay		; 100us delat
Ser_Loop	rrf		TX_Byte,f		; Rotate through C to test for next bit
		btfss		STATUS,C		; If C is off
		bcf		Tx			; TX Off
		btfsc		STATUS,C		; If C is ON
		bsf		Tx			; TX ON
		call		Bit_Delay		; 100us delay
		decfsz		Bit_Cntr,f		; test if all 8 bits are done
		goto		Ser_Loop		; No, next bit
		bsf		Tx			; Yes, switch On TX (Stop Bit)
		call		Bit_Delay		; 100us
		bcf		Tx			; Switch OFF TX
		bcf		DEn			; Switch OFF DataEnable
		return

; ============================================================================= ;
; Serial routines - RX - RX - RX - RX - RX - RX - RX - RX - RX - RX - RX - RX
; ============================================================================= ;
;
; --- Serial Receive ---------------------------------------------------------- ;
; Receive data routine
; Should no data be received within a period of 1ms (time to transmit one byte)
; (Timer1 timeout) the routine is returned
; ----------------------------------------------------------------------------- ;
; (Receive routine)
;      St  0   1   2   3   4   5   6   7   Sp
;    +---+---+---+---+---+---+---+---+---+---+
;    | 1 |/X/|/X/|/X/|/X/|/X/|/X/|/X/|/X/| 1 |
;----+   +---+---+---+---+---+---+---+---+   +---------
;      |   |   |   |   |   |   |   |   |   |
;      |   |   |   |   |   |   |   |   |   +- Stop Bit Logic 1
;      |   |   |   |   |   |   |   |   +----- MSB Nibble <7>
;      |   |   |   |   |   |   |   +---------     Nibble <6>
;      |   |   |   |   |   |   +-------------     Nibble <5>
;      |   |   |   |   |   +-----------------     Nibble <4>
;      |   |   |   |   +---------------------     Nibble <3>
;      |   |   |   +-------------------------     Nibble <2>
;      |   |   +-----------------------------     Nibble <1>
;      |   +--------------------------------- LSB Nibble <0>
;      +------------------------------------- Start Bit Logic 1
;
; ----------------------------------------------------------------------------- ;

RX_232:		
		call		EnTmr1			; Enable Timer1 
RxWait:		btfss		Tmr1Flag		; Test if Timer1 flag countout has occurred
		goto		TstRxTx			; Test if RxTx is set?
		bcf		Tmr1Flag		; Reset Timer1 Overflow tag
		call		DisTmr1			; Disable Timer1
		return					;
TstRxTx:	btfss		Rx			; Test if Rx is set <1>
		goto		RxWait			; NO
; ----------------------------------------------------- ;
		call		Start_Delay		; do half bit time delay (53us)
		btfss		Rx			; check still in start bit
		goto		RxWait			; NO
; ----------------------------------------------------- ;
; At this point we confirm that the start bit (0) has
; been received and we can disable Timer1 watchdog	
; ----------------------------------------------------- ;
		call		DisTmr1			; YES - Disable Timer1
		movlw		0x08			; set up to read 8 bits
		movwf		Bit_Cntr		;
		clrf		RX_Byte			;
Next_RcvBit	call		Bit_Delay		; 101us Delay
		btfss		Rx			; Test Rx 
		bcf		STATUS,C		;
		btfsc		Rx			; Test Rx 	
		bsf		STATUS,C		;
		rrf		RX_Byte,f		;
		decfsz		Bit_Cntr,f		; test if all done
		goto		Next_RcvBit		;
		call		Bit_Delay		; 101us Delay
		return					;

; --- Decode ------------------------------------------------------------------ ;
; Received data to be displayed on LED
; Nodes 1 to 8 each have one Input and one output signal
; The master transmits a request (On or Off) and then that drives the Slave's
; output (relay or LED) according to the requirements from the master. 
; The result of the aciton taken by the slave's (feedback) is then sent back to 
; the master for confirmation.
; Receiving data is received in RX_Byte, where bits 0-2 is used to decode the node
; address of the slave, and the last bit<7> holds the data value for that node
; ----------------------------------------------------------------------------- ;
Decode:
GetNode:	movf		RX_Byte,W		;
		andlw		0x0F			; filter out node address
		movwf		RxNodeNo		; Place into RxNodeNo reg
DecodeRx:	clrf		Temp_Byte		;
		movf		RxNodeNo,w		;
		andlw		7			; w = index, 0..7
		addwf		PCL,f			; 
							;
		goto		N01			; If Node address = 0x00
		goto		N02			; If Node address = 0x01
		goto		N03			; If Node address = 0x02
		goto		N04			; If Node address = 0x03
		goto		N05			; If Node address = 0x04
		goto		N06			; If Node address = 0x05
		goto		N07			; If Node address = 0x06
		goto		N08			; If Node address = 0x07
		return					; No valid node
; ----------------------------------------------------- ;
N01:		btfss		RX_Byte,7		; Test if Rx Data is set
		bcf		PORTC,0			; Clear the LED
		btfsc		RX_Byte,7		; Test if Rx Data is reset
		bsf		PORTC,0			; Set the LED
		return					;
; ----------------------------------------------------- ;
N02:		btfss		RX_Byte,7		; Test if Rx Data is set
		bcf		PORTC,1			; Clear the LED
		btfsc		RX_Byte,7		; Test if Rx Data is reset
		bsf		PORTC,1			; Set the LED
		return					;
; ----------------------------------------------------- ;
N03:		btfss		RX_Byte,7		; Test if Rx Data is set
		bcf		PORTC,2			; Clear the LED
		btfsc		RX_Byte,7		; Test if Rx Data is reset
		bsf		PORTC,2			; Set the LED
		return					;
; ----------------------------------------------------- ;
N04:		btfss		RX_Byte,7		; Test if Rx Data is set
		bcf		PORTC,3			; Clear the LED
		btfsc		RX_Byte,7		; Test if Rx Data is reset
		bsf		PORTC,3			; Set the LED
		return					;
; ----------------------------------------------------- ;
N05:		btfss		RX_Byte,7		; Test if Rx Data is set
		bcf		PORTC,4			; Clear the LED
		btfsc		RX_Byte,7		; Test if Rx Data is reset
		bsf		PORTC,4			; Set the LED
		return					;
; ----------------------------------------------------- ;
N06:		btfss		RX_Byte,7		; Test if Rx Data is set
		bcf		PORTC,5			; Clear the LED
		btfsc		RX_Byte,7		; Test if Rx Data is reset
		bsf		PORTC,5			; Set the LED
		return					;
; ----------------------------------------------------- ;
N07:		btfss		RX_Byte,7		; Test if Rx Data is set
		bcf		PORTC,6			; Clear the LED
		btfsc		RX_Byte,7		; Test if Rx Data is reset
		bsf		PORTC,6			; Set the LED
		return					;
; ----------------------------------------------------- ;
N08:		btfss		RX_Byte,7		; Test if Rx Data is set
		bcf		PORTC,7			; Clear the LED
		btfsc		RX_Byte,7		; Test if Rx Data is reset
		bsf		PORTC,7			; Set the LED
		return					;
; ----------------------------------------------------- ;

; ============================================================================= ;
; SET ANALOG/DIGITAL INPUTS PORT A 			;
; ------------------------------------------------------------------------------ 
Start:
; ------------------------------------------------------------------------------
; INITIALIZE PORTS					; 
; ------------------------------------------------------------------------------
		bsf		STATUS,RP0		; | B1
		movlw		0x06			; (16F876)
		movwf		ADCON1			; (16F876)
		movlw		b'00000001'		; PortA<0> Inputs
		movwf		TRISA			; PortA<1,2,3,4,5> Outputs 
		movlw		b'11111111'		; All PortB pins inputs
		movwf		TRISB			; Node command switch
		movlw		b'00000000'		; All outputs
		movwf		TRISC			; LED indications
		movlw		0x00			; Weak pull-ups enabled
		movwf		OPTION_REG		;
		bcf		STATUS,RP0		; | B0
; ----------------------------------------------------->;
; Indirect register resetting
; ----------------------------------------------------->;
		movlw		0x20			; Put first address in for
		movwf		FSR			; FSR pointer to go to
Next:		clrf		INDF			; Clear the register at that address
		incf		FSR,f			; increment FSR with 1
		movf		FSR,w			; Put FSR in w
		sublw		0x30			; Clear all registers from 0x20-0x30
		btfss		STATUS,Z		; Test if all done
		goto		Next			; No
; ------------> Variable node address			;
		movlw 		0xFF			; Set node no to 255 to overflow
		movwf		NodeNo			; on next execution
		return					; return INIT call
; ==============================================================================

; --- Enable Timer 1 funtion -------------------------------------------------- ;
EnTmr1:							;
; -----------------------------------------------------------------------------	;
		bcf		PIR1,TMR1IF		; Reset Timer overflow flag (When interrupt occurred)
		movlw		high(Val)		;
		movwf		TMR1H			; write value into hi_byte timer register
		movlw		low(Val)		; 
		movwf		TMR1L			; write value into lo_byte timer register
		movlw		B'00000001'		; Pre-scaler set to 1, LP osc DISABLED, Start timer '0000 0001'
		movwf		T1CON			; Move into Timer1 Control Register
		bsf     	STATUS,RP0		; | B1
		bsf		PIE1,TMR1IE		; enable Timer1 Overflow interrupt
		bcf     	STATUS,RP0		; | B0
		bsf		INTCON,GIE		; Global interrupt enable
		bsf		INTCON,PEIE		; Enable Periphiral interrupt
		return					;

; --- Disable Timer 1 funtion ------------------------------------------------- ;
DisTmr1:						;
; -----------------------------------------------------------------------------	;
		bsf     	STATUS,RP0		; | B1
		bcf		PIE1,TMR1IE		; Disable Timer1 Overflow interrupt		
		bcf     	STATUS,RP0		; | B0
		return					; 

; ----------------------------------------------------- ;
; Generates 50us Delay
; ----------------------------------------------------- ;
Start_Delay	movlw		d'010'			; Was 0x0C
		movwf		Delay_Count		;
Start_Wait	nop					;
		decfsz		Delay_Count , f		;
		goto		Start_Wait		;
		return					;
							;
; ------------------------------------------------------;
;
; ------------------------------------------------------;
Bit_Delay	movlw		d'022'			; was 0x18
		movwf		Delay_Count		;
Bit_Wait	nop					;
		decfsz		Delay_Count , f		;
		goto		Bit_Wait		;
		return					;
							;
; ------------------------------------------------------
;Delay routines
; ------------------------------------------------------
							;
Delay255	movlw		0xff			;delay 255 mS
		goto		d0			;
Delay100	movlw		d'100'			;delay 100mS
		goto		d0			;
Delay50		movlw		d'50'			;delay 50mS
		goto		d0			;
Delay20		movlw		d'20'			;delay 20mS
		goto		d0			;
Delay10		movlw		d'10'			;delay 10mS
		goto		d0			;
Delay1		movlw		d'1'			;delay 1mS
		goto		d0			;
Delay5		movlw		0x05			;delay 5.000 ms (4 MHz clock)
d0		movwf		count1			;
d1		movlw		0xC7			;
		movwf		counta			;
		movlw		0x01			;
		movwf		countb			;
Delay_0		decfsz		counta, f		;
		goto		$+2			;
		decfsz		countb, f		;
		goto		Delay_0			;
		decfsz		count1	,f		;
		goto		d1			;
		retlw		0x00			;
							;
; ===================================================== ;
; end of Miscellaneous routines
; ------------------------------------------------------;

		end


CREDITS
I based the Rx/Tx routines on code from Nigel Goodwin's page
http://www.winpicprog.co.uk/pic_tutorial7.htm
http://www.winpicprog.co.uk/